@@ -12,6 +12,14 @@
 #include <linux/sizes.h>
 #include "vfsmod.h"
 
+struct sf_handle {
+	u64 handle;
+	u32 root;
+	u32 access_flags;
+	struct kref refcount;
+	struct list_head head;
+};
+
 /**
  * Read from a regular file.
  * Return: The number of bytes read on success, negative errno value otherwise
@@ -23,8 +31,7 @@
 static ssize_t sf_reg_read(struct file *file, char __user *buf, size_t size,
 			   loff_t *off)
 {
-	struct sf_glob_info *sf_g = GET_GLOB_INFO(file_inode(file)->i_sb);
-	struct sf_reg_info *sf_r = file->private_data;
+	struct sf_handle *sf_handle = file->private_data;
 	u64 pos = *off;
 	u32 nread;
 	int err;
@@ -37,7 +44,7 @@ static ssize_t sf_reg_read(struct file *file, char __user *buf, size_t size,
 	else
 		nread = size;
 
-	err = vboxsf_read(sf_g->root, sf_r->handle, pos, &nread,
+	err = vboxsf_read(sf_handle->root, sf_handle->handle, pos, &nread,
 			  (uintptr_t)buf, true);
 	if (err)
 		return err;
@@ -59,8 +66,7 @@ static ssize_t sf_reg_write(struct file *file, const char __user *buf,
 {
 	struct inode *inode = file_inode(file);
 	struct sf_inode_info *sf_i = GET_INODE_INFO(inode);
-	struct sf_glob_info *sf_g = GET_GLOB_INFO(inode->i_sb);
-	struct sf_reg_info *sf_r = file->private_data;
+	struct sf_handle *sf_handle = file->private_data;
 	u32 nwritten;
 	u64 pos;
 	int err;
@@ -83,7 +89,7 @@ static ssize_t sf_reg_write(struct file *file, const char __user *buf,
 	if (err)
 		return err;
 
-	err = vboxsf_write(sf_g->root, sf_r->handle, pos, &nwritten,
+	err = vboxsf_write(sf_handle->root, sf_handle->handle, pos, &nwritten,
 			   (uintptr_t)buf, true);
 	if (err)
 		return err;
@@ -112,11 +118,12 @@ static int sf_reg_open(struct inode *inode, struct file *file)
 {
 	struct sf_inode_info *sf_i = GET_INODE_INFO(inode);
 	struct shfl_createparms params = {};
-	struct sf_reg_info *sf_r;
+	struct sf_handle *sf_handle;
+	u32 access_flags = 0;
 	int err;
 
-	sf_r = kmalloc(sizeof(*sf_r), GFP_KERNEL);
-	if (!sf_r)
+	sf_handle = kmalloc(sizeof(*sf_handle), GFP_KERNEL);
+	if (!sf_handle)
 		return -ENOMEM;
 
 	/*
@@ -146,15 +153,15 @@ static int sf_reg_open(struct inode *inode, struct file *file)
 
 	switch (file->f_flags & O_ACCMODE) {
 	case O_RDONLY:
-		params.create_flags |= SHFL_CF_ACCESS_READ;
+		access_flags |= SHFL_CF_ACCESS_READ;
 		break;
 
 	case O_WRONLY:
-		params.create_flags |= SHFL_CF_ACCESS_WRITE;
+		access_flags |= SHFL_CF_ACCESS_WRITE;
 		break;
 
 	case O_RDWR:
-		params.create_flags |= SHFL_CF_ACCESS_READWRITE;
+		access_flags |= SHFL_CF_ACCESS_READWRITE;
 		break;
 
 	default:
@@ -162,26 +169,45 @@ static int sf_reg_open(struct inode *inode, struct file *file)
 	}
 
 	if (file->f_flags & O_APPEND)
-		params.create_flags |= SHFL_CF_ACCESS_APPEND;
+		access_flags |= SHFL_CF_ACCESS_APPEND;
 
+	params.create_flags |= access_flags;
 	params.info.attr.mode = inode->i_mode;
 
 	err = vboxsf_create_at_dentry(file_dentry(file), &params);
 	if (err == 0 && params.handle == SHFL_HANDLE_NIL)
 		err = (params.result == SHFL_FILE_EXISTS) ? -EEXIST : -ENOENT;
 	if (err) {
-		kfree(sf_r);
+		kfree(sf_handle);
 		return err;
 	}
 
 	/* the host may have given us different attr then requested */
 	sf_i->force_restat = 1;
-	sf_r->handle = params.handle;
-	sf_i->file = file;
-	file->private_data = sf_r;
+
+	/* init our handle struct and add it to the inode's handles list */
+	sf_handle->handle = params.handle;
+	sf_handle->root = GET_GLOB_INFO(inode->i_sb)->root;
+	sf_handle->access_flags = access_flags;
+	kref_init(&sf_handle->refcount);
+
+	mutex_lock(&sf_i->handle_list_mutex);
+	list_add(&sf_handle->head, &sf_i->handle_list);
+	mutex_unlock(&sf_i->handle_list_mutex);
+
+	file->private_data = sf_handle;
 	return 0;
 }
 
+static void sf_handle_release(struct kref *refcount)
+{
+	struct sf_handle *sf_handle = container_of(refcount, struct sf_handle,
+						   refcount);
+
+	vboxsf_close(sf_handle->root, sf_handle->handle);
+	kfree(sf_handle);
+}
+
 /**
  * Close a regular file.
  * Return: 0 or negative errno value.
@@ -190,29 +216,67 @@ static int sf_reg_open(struct inode *inode, struct file *file)
  */
 static int sf_reg_release(struct inode *inode, struct file *file)
 {
-	struct sf_reg_info *sf_r;
-	struct sf_glob_info *sf_g;
 	struct sf_inode_info *sf_i = GET_INODE_INFO(inode);
-
-	sf_g = GET_GLOB_INFO(inode->i_sb);
-	sf_r = file->private_data;
+	struct sf_handle *sf_handle = file->private_data;
 
 	filemap_write_and_wait(inode->i_mapping);
 
-	vboxsf_close(sf_g->root, sf_r->handle);
+	mutex_lock(&sf_i->handle_list_mutex);
+	list_del(&sf_handle->head);
+	mutex_unlock(&sf_i->handle_list_mutex);
 
-	kfree(sf_r);
-	sf_i->file = NULL;
+	kref_put(&sf_handle->refcount, sf_handle_release);
 	file->private_data = NULL;
 	return 0;
 }
 
+/*
+ * Write back dirty pages now, because there may not be any suitable
+ * open files later
+ */
+static void sf_vma_close(struct vm_area_struct *vma)
+{
+	filemap_write_and_wait(vma->vm_file->f_mapping);
+}
+
+static vm_fault_t sf_page_mkwrite(struct vm_fault *vmf)
+{
+	struct page *page = vmf->page;
+	struct inode *inode = file_inode(vmf->vma->vm_file);
+
+	lock_page(page);
+	if (page->mapping != inode->i_mapping) {
+		unlock_page(page);
+		return VM_FAULT_NOPAGE;
+	}
+
+	return VM_FAULT_LOCKED;
+}
+
+static const struct vm_operations_struct sf_file_vm_ops = {
+	.close		= sf_vma_close,
+	.fault		= filemap_fault,
+	.map_pages	= filemap_map_pages,
+	.page_mkwrite	= sf_page_mkwrite,
+};
+
+static int sf_reg_mmap(struct file *file, struct vm_area_struct *vma)
+{
+	int err;
+
+	err = generic_file_mmap(file, vma);
+	if (!err)
+		vma->vm_ops = &sf_file_vm_ops;
+
+	return err;
+}
+
 const struct file_operations vboxsf_reg_fops = {
 	.read = sf_reg_read,
 	.open = sf_reg_open,
 	.write = sf_reg_write,
 	.release = sf_reg_release,
-	.mmap = generic_file_mmap,
+	.mmap = sf_reg_mmap,
 	.splice_read = generic_file_splice_read,
 	.read_iter = generic_file_read_iter,
 	.write_iter = generic_file_write_iter,
@@ -227,8 +291,7 @@ const struct inode_operations vboxsf_reg_iops = {
 
 static int sf_readpage(struct file *file, struct page *page)
 {
-	struct sf_glob_info *sf_g = GET_GLOB_INFO(file_inode(file)->i_sb);
-	struct sf_reg_info *sf_r = file->private_data;
+	struct sf_handle *sf_handle = file->private_data;
 	loff_t off = page_offset(page);
 	u32 nread = PAGE_SIZE;
 	u8 *buf;
@@ -236,7 +299,7 @@ static int sf_readpage(struct file *file, struct page *page)
 
 	buf = kmap(page);
 
-	err = vboxsf_read(sf_g->root, sf_r->handle, off, &nread,
+	err = vboxsf_read(sf_handle->root, sf_handle->handle, off, &nread,
 			  (uintptr_t)buf, false);
 	if (err == 0) {
 		memset(&buf[nread], 0, PAGE_SIZE - nread);
@@ -251,12 +314,29 @@ static int sf_readpage(struct file *file, struct page *page)
 	return err;
 }
 
+static struct sf_handle *sf_get_writeable_handle(struct sf_inode_info *sf_i)
+{
+	struct sf_handle *h, *sf_handle = NULL;
+
+	mutex_lock(&sf_i->handle_list_mutex);
+	list_for_each_entry(h, &sf_i->handle_list, head) {
+		if (h->access_flags == SHFL_CF_ACCESS_WRITE ||
+		    h->access_flags == SHFL_CF_ACCESS_READWRITE) {
+			kref_get(&h->refcount);
+			sf_handle = h;
+			break;
+		}
+	}
+	mutex_unlock(&sf_i->handle_list_mutex);
+
+	return sf_handle;
+}
+
 static int sf_writepage(struct page *page, struct writeback_control *wbc)
 {
 	struct inode *inode = page->mapping->host;
-	struct sf_glob_info *sf_g = GET_GLOB_INFO(inode->i_sb);
 	struct sf_inode_info *sf_i = GET_INODE_INFO(inode);
-	struct sf_reg_info *sf_r = sf_i->file->private_data;
+	struct sf_handle *sf_handle;
 	loff_t off = page_offset(page);
 	loff_t size = i_size_read(inode);
 	u32 nwrite = PAGE_SIZE;
@@ -266,11 +346,17 @@ static int sf_writepage(struct page *page, struct writeback_control *wbc)
 	if (off + PAGE_SIZE > size)
 		nwrite = size & ~PAGE_MASK;
 
+	sf_handle = sf_get_writeable_handle(sf_i);
+	if (!sf_handle)
+		return -EBADF;
+
 	buf = kmap(page);
-	err = vboxsf_write(sf_g->root, sf_r->handle, off, &nwrite,
+	err = vboxsf_write(sf_handle->root, sf_handle->handle, off, &nwrite,
 			   (uintptr_t)buf, false);
 	kunmap(page);
 
+	kref_put(&sf_handle->refcount, sf_handle_release);
+
 	if (err == 0) {
 		ClearPageError(page);
 		/* mtime changed */
@@ -288,15 +374,14 @@ static int sf_write_end(struct file *file, struct address_space *mapping,
 			struct page *page, void *fsdata)
 {
 	struct inode *inode = mapping->host;
-	struct sf_glob_info *sf_g = GET_GLOB_INFO(inode->i_sb);
-	struct sf_reg_info *sf_r = file->private_data;
+	struct sf_handle *sf_handle = file->private_data;
 	unsigned int from = pos & ~PAGE_MASK;
 	u32 nwritten = len;
 	u8 *buf;
 	int err;
 
 	buf = kmap(page);
-	err = vboxsf_write(sf_g->root, sf_r->handle, pos, &nwritten,
+	err = vboxsf_write(sf_handle->root, sf_handle->handle, pos, &nwritten,
 			   (uintptr_t)buf + from, false);
 	kunmap(page);
 
